@@ -57,6 +57,8 @@ const (
 	ProcessingPhasePause ProcessingPhase = "Pause"
 	// ProcessingPhaseError is the phase in which we encountered an error and need to exit ungracefully.
 	ProcessingPhaseError ProcessingPhase = "Error"
+	// ProcessingPhaseMergeDelta is the phase in a multi-stage import where a delta image downloaded to scratch is applied to the base image
+	ProcessingPhaseMergeDelta ProcessingPhase = "MergeDelta"
 )
 
 // ValidationSizeError is an error indication size validation failure.
@@ -129,6 +131,10 @@ func NewDataProcessor(dataSource DataSourceInterface, dataFile, dataDir, scratch
 	if isVddk {
 		needsDataCleanup = !vddkSource.IsDeltaCopy()
 	}
+	imageioSource, isImageio := dataSource.(*ImageioDataSource)
+	if isImageio {
+		needsDataCleanup = !imageioSource.IsDeltaCopy()
+	}
 	dp := &DataProcessor{
 		currentPhase:       ProcessingPhaseInfo,
 		source:             dataSource,
@@ -221,6 +227,11 @@ func (dp *DataProcessor) ProcessDataWithPause() error {
 			if err != nil {
 				err = errors.Wrap(err, "Unable to resize disk image to requested size")
 			}
+		case ProcessingPhaseMergeDelta:
+			dp.currentPhase, err = dp.merge()
+			if err != nil {
+				err = errors.Wrap(err, "Unable to apply delta to base image")
+			}
 		default:
 			return errors.Errorf("Unknown processing phase %s", dp.currentPhase)
 		}
@@ -313,6 +324,11 @@ func ResizeImage(dataFile, imageSize string, totalTargetSpace int64, preallocati
 			klog.V(1).Infof("No need to resize image. Requested size: %s, Image size: %d.\n", imageSize, info.VirtualSize)
 			return nil
 		}
+		// Check if calculated size is < imageSize, and return error if so.
+		if currentImageSizeQuantity.Cmp(minSizeQuantity) == 1 {
+			klog.V(1).Infof("Calculated new size is < than current size, not resizing: requested size %s, virtual size: %d.\n", minSizeQuantity.String(), info.VirtualSize)
+			return nil
+		}
 		klog.V(1).Infof("Expanding image size to: %s\n", minSizeQuantity.String())
 		return qemuOperations.Resize(dataFile, minSizeQuantity, preallocation)
 	}
@@ -356,14 +372,21 @@ func (dp *DataProcessor) PreallocationApplied() bool {
 }
 
 func (dp *DataProcessor) getUsableSpace() int64 {
-	return GetUsableSpace(dp.filesystemOverhead, dp.availableSpace)
+	return util.GetUsableSpace(dp.filesystemOverhead, dp.availableSpace)
 }
 
-// GetUsableSpace calculates space to use taking file system overhead into account
-func GetUsableSpace(filesystemOverhead float64, availableSpace int64) int64 {
-	spaceWithOverhead := int64((1 - filesystemOverhead) * float64(availableSpace))
-	// qemu-img will round up, making us use more than the usable space.
-	// This later conflicts with image size validation.
-	qemuImgCorrection := util.RoundDown(spaceWithOverhead, util.DefaultAlignBlockSize)
-	return qemuImgCorrection
+// Rebase and commit a delta image to its backing file
+func (dp *DataProcessor) merge() (ProcessingPhase, error) {
+	klog.V(1).Info("Merging QCOW to base image.")
+	imageURL := dp.source.GetURL()
+	if imageURL == nil {
+		return ProcessingPhaseError, errors.New("bad URL in data source")
+	}
+	if err := qemuOperations.Rebase(dp.dataFile, imageURL.String()); err != nil {
+		return ProcessingPhaseError, errors.Wrap(err, "error rebasing image")
+	}
+	if err := qemuOperations.Commit(imageURL.String()); err != nil {
+		return ProcessingPhaseError, errors.Wrap(err, "error committing image")
+	}
+	return ProcessingPhaseComplete, nil
 }
