@@ -5,6 +5,8 @@ import (
 	"fmt"
 	"time"
 
+	"github.com/onsi/gomega"
+
 	corev1 "k8s.io/api/core/v1"
 	k8sv1 "k8s.io/api/core/v1"
 	apierrs "k8s.io/apimachinery/pkg/api/errors"
@@ -13,7 +15,7 @@ import (
 	"k8s.io/apimachinery/pkg/runtime/schema"
 	"k8s.io/apimachinery/pkg/util/wait"
 
-	cdiv1 "kubevirt.io/containerized-data-importer/pkg/apis/core/v1beta1"
+	cdiv1 "kubevirt.io/containerized-data-importer-api/pkg/apis/core/v1beta1"
 	cdiclientset "kubevirt.io/containerized-data-importer/pkg/client/clientset/versioned"
 )
 
@@ -33,12 +35,24 @@ const (
 	TinyCoreIsoRegistryURL = "docker://cdi-docker-registry-host.%s/tinycoreqcow2"
 	//TinyCoreIsoRegistryProxyURL provides a test url for the tinycore.qcow2 image wrapped in docker container available through rate-limiting proxy
 	TinyCoreIsoRegistryProxyURL = "docker://cdi-file-host.%s:83/tinycoreqcow2"
+	// TinyCoreIsoAuthURL provides a tinyCore ISO from a URL that requires basic authentication
+	TinyCoreIsoAuthURL = "http://cdi-file-host.%s:81/tinyCore.iso"
 	// HTTPSTinyCoreIsoURL provides a test (https) url for the tineyCore iso image
 	HTTPSTinyCoreIsoURL = "https://cdi-file-host.%s/tinyCore.iso"
 	// HTTPSTinyCoreQcow2URL provides a test (https) url for the tineyCore qcow2 image
 	HTTPSTinyCoreQcow2URL = "https://cdi-file-host.%s/tinyCore.qcow2"
 	// TinyCoreQcow2URLRateLimit provides a test url for the tineyCore qcow2 image via rate-limiting proxy
 	TinyCoreQcow2URLRateLimit = "http://cdi-file-host.%s:82/tinyCore.qcow2"
+	// TinyCoreQcow2GzURLRateLimit provides a test url for the tineyCore qcow2.gz image via rate-limiting proxy
+	TinyCoreQcow2GzURLRateLimit = "http://cdi-file-host.%s:82/tinyCore.qcow2.gz"
+	// HTTPSTinyCoreVmdkURL provides a test url for the tineyCore qcow2 image
+	HTTPSTinyCoreVmdkURL = "https://cdi-file-host.%s/tinyCore.vmdk"
+	// HTTPSTinyCoreVdiURL provides a test url for the tineyCore qcow2 image
+	HTTPSTinyCoreVdiURL = "https://cdi-file-host.%s/tinyCore.vdi"
+	// HTTPSTinyCoreVhdURL provides a test url for the tineyCore qcow2 image
+	HTTPSTinyCoreVhdURL = "https://cdi-file-host.%s/tinyCore.vhd"
+	// HTTPSTinyCoreVhdxURL provides a test url for the tineyCore qcow2 image
+	HTTPSTinyCoreVhdxURL = "https://cdi-file-host.%s/tinyCore.vhdx"
 	// InvalidQcowImagesURL provides a test url for invalid qcow images
 	InvalidQcowImagesURL = "http://cdi-file-host.%s/invalid_qcow_images/"
 	// LargeVirtualDiskQcow provides a test url for a cirros image with a large virtual size, in qcow2 format
@@ -55,15 +69,23 @@ const (
 	CirrosURL = "http://cdi-file-host.%s/cirros-qcow2.img"
 	// ImageioURL provides URL of oVirt engine hosting imageio
 	ImageioURL = "https://imageio.%s:12346/ovirt-engine/api"
+	// ImageioRootURL provides the base path to fakeovirt, for inventory modifications
+	ImageioRootURL = "https://imageio.%s:12346"
+	// ImageioImageURL provides the base path to imageiotest, for test images
+	ImageioImageURL = "https://imageio.%s:12345"
 	// VcenterURL provides URL of vCenter/ESX simulator
 	VcenterURL = "https://vcenter.%s:8989/sdk"
+	// TrustedRegistryURL provides the base path to trusted registry test url for the tinycore.iso image wrapped in docker container
+	TrustedRegistryURL = "docker://%s/cdi-func-test-tinycore"
+	// TrustedRegistryIS provides the base path to trusted registry test fake imagestream for the tinycore.iso image wrapped in docker container
+	TrustedRegistryIS = "%s/cdi-func-test-tinycore"
 
+	// MD5PrefixSize is the number of bytes used by the MD5 constants below
+	MD5PrefixSize = int64(100000)
 	// TinyCoreMD5 is the MD5 hash of first 100k bytes of tinyCore image
 	TinyCoreMD5 = "3710416a680523c7d07538cb1026c60c"
 	// TinyCoreTarMD5 is the MD5 hash of first 100k bytes of tinyCore tar image
 	TinyCoreTarMD5 = "aec1a39d753b4b7cc81ee02bc625a342"
-	// TinyCoreBlockMD5 is the MD5 hash of first 100k bytes of tinyCore image on block device
-	TinyCoreBlockMD5 = "d41d8cd98f00b204e9800998ecf8427e"
 	// ImageioMD5 is the MD5 hash of first 100k bytes of imageio image
 	ImageioMD5 = "91150be031835ccfac458744da57d4f6"
 	// VcenterMD5 is the MD5 hash of first 100k bytes of Vcenter image
@@ -105,19 +127,120 @@ func NewCloningDataVolume(dataVolumeName, size string, sourcePvc *k8sv1.Persiste
 	return NewDataVolumeForImageCloning(dataVolumeName, size, sourcePvc.Namespace, sourcePvc.Name, sourcePvc.Spec.StorageClassName, sourcePvc.Spec.VolumeMode)
 }
 
-// NewDataVolumeWithHTTPImport initializes a DataVolume struct with HTTP annotations
-func NewDataVolumeWithHTTPImport(dataVolumeName string, size string, httpURL string) *cdiv1.DataVolume {
+// NewDataVolumeWithSourceRef initializes a DataVolume struct with DataSource SourceRef
+func NewDataVolumeWithSourceRef(dataVolumeName string, size, sourceRefNamespace, sourceRefName string) *cdiv1.DataVolume {
+	claimSpec := &k8sv1.PersistentVolumeClaimSpec{
+		AccessModes: []k8sv1.PersistentVolumeAccessMode{k8sv1.ReadWriteOnce},
+		Resources: k8sv1.ResourceRequirements{
+			Requests: k8sv1.ResourceList{
+				k8sv1.ResourceStorage: resource.MustParse(size),
+			},
+		},
+	}
+
 	return &cdiv1.DataVolume{
 		ObjectMeta: metav1.ObjectMeta{
 			Name:        dataVolumeName,
 			Annotations: map[string]string{},
 		},
 		Spec: cdiv1.DataVolumeSpec{
-			Source: cdiv1.DataVolumeSource{
+			SourceRef: &cdiv1.DataVolumeSourceRef{
+				Kind:      cdiv1.DataVolumeDataSource,
+				Namespace: &sourceRefNamespace,
+				Name:      sourceRefName,
+			},
+			PVC: claimSpec,
+		},
+	}
+}
+
+// NewDataSource initializes a DataSource struct with PVC source
+func NewDataSource(dataSourceName, dataSourceNamespace, pvcName, pvcNamespace string) *cdiv1.DataSource {
+	return &cdiv1.DataSource{
+		ObjectMeta: metav1.ObjectMeta{
+			Name:      dataSourceName,
+			Namespace: dataSourceNamespace,
+		},
+		Spec: cdiv1.DataSourceSpec{
+			Source: cdiv1.DataSourceSource{
+				PVC: &cdiv1.DataVolumeSourcePVC{
+					Name:      pvcName,
+					Namespace: pvcNamespace,
+				},
+			},
+		},
+	}
+}
+
+// NewDataVolumeWithHTTPImport initializes a DataVolume struct with HTTP annotations
+func NewDataVolumeWithHTTPImport(dataVolumeName string, size string, httpURL string) *cdiv1.DataVolume {
+	claimSpec := &k8sv1.PersistentVolumeClaimSpec{
+		AccessModes: []k8sv1.PersistentVolumeAccessMode{k8sv1.ReadWriteOnce},
+		Resources: k8sv1.ResourceRequirements{
+			Requests: k8sv1.ResourceList{
+				k8sv1.ResourceStorage: resource.MustParse(size),
+			},
+		},
+	}
+
+	return &cdiv1.DataVolume{
+		ObjectMeta: metav1.ObjectMeta{
+			Name:        dataVolumeName,
+			Annotations: map[string]string{},
+		},
+		Spec: cdiv1.DataVolumeSpec{
+			Source: &cdiv1.DataVolumeSource{
 				HTTP: &cdiv1.DataVolumeSourceHTTP{
 					URL: httpURL,
 				},
 			},
+			PVC: claimSpec,
+		},
+	}
+}
+
+// NewDataVolumeWithHTTPImportAndStorageSpec initializes a DataVolume struct with HTTP annotations
+func NewDataVolumeWithHTTPImportAndStorageSpec(dataVolumeName string, size string, httpURL string) *cdiv1.DataVolume {
+	storageSpec := &cdiv1.StorageSpec{
+		AccessModes: []k8sv1.PersistentVolumeAccessMode{k8sv1.ReadWriteOnce},
+		Resources: k8sv1.ResourceRequirements{
+			Requests: k8sv1.ResourceList{
+				k8sv1.ResourceStorage: resource.MustParse(size),
+			},
+		},
+	}
+
+	return &cdiv1.DataVolume{
+		ObjectMeta: metav1.ObjectMeta{
+			Name:        dataVolumeName,
+			Annotations: map[string]string{},
+		},
+		Spec: cdiv1.DataVolumeSpec{
+			Source: &cdiv1.DataVolumeSource{
+				HTTP: &cdiv1.DataVolumeSourceHTTP{
+					URL: httpURL,
+				},
+			},
+			Storage: storageSpec,
+		},
+	}
+}
+
+// NewDataVolumeWithImageioImport initializes a DataVolume struct with Imageio annotations
+func NewDataVolumeWithImageioImport(dataVolumeName string, size string, httpURL string, secret string, configMap string, diskID string) *cdiv1.DataVolume {
+	return &cdiv1.DataVolume{
+		ObjectMeta: metav1.ObjectMeta{
+			Name: dataVolumeName,
+		},
+		Spec: cdiv1.DataVolumeSpec{
+			Source: &cdiv1.DataVolumeSource{
+				Imageio: &cdiv1.DataVolumeSourceImageIO{
+					URL:           httpURL,
+					SecretRef:     secret,
+					CertConfigMap: configMap,
+					DiskID:        diskID,
+				},
+			},
 			PVC: &k8sv1.PersistentVolumeClaimSpec{
 				AccessModes: []k8sv1.PersistentVolumeAccessMode{k8sv1.ReadWriteOnce},
 				Resources: k8sv1.ResourceRequirements{
@@ -130,14 +253,14 @@ func NewDataVolumeWithHTTPImport(dataVolumeName string, size string, httpURL str
 	}
 }
 
-// NewDataVolumeWithImageioImport initializes a DataVolume struct with Imageio annotations
-func NewDataVolumeWithImageioImport(dataVolumeName string, size string, httpURL string, secret string, configMap string, diskID string) *cdiv1.DataVolume {
+// NewDataVolumeWithImageioWarmImport initializes a DataVolume struct for a multi-stage import from oVirt snapshots
+func NewDataVolumeWithImageioWarmImport(dataVolumeName string, size string, httpURL string, secret string, configMap string, diskID string, checkpoints []cdiv1.DataVolumeCheckpoint, finalCheckpoint bool) *cdiv1.DataVolume {
 	return &cdiv1.DataVolume{
 		ObjectMeta: metav1.ObjectMeta{
 			Name: dataVolumeName,
 		},
 		Spec: cdiv1.DataVolumeSpec{
-			Source: cdiv1.DataVolumeSource{
+			Source: &cdiv1.DataVolumeSource{
 				Imageio: &cdiv1.DataVolumeSourceImageIO{
 					URL:           httpURL,
 					SecretRef:     secret,
@@ -145,6 +268,8 @@ func NewDataVolumeWithImageioImport(dataVolumeName string, size string, httpURL
 					DiskID:        diskID,
 				},
 			},
+			FinalCheckpoint: finalCheckpoint,
+			Checkpoints:     checkpoints,
 			PVC: &k8sv1.PersistentVolumeClaimSpec{
 				AccessModes: []k8sv1.PersistentVolumeAccessMode{k8sv1.ReadWriteOnce},
 				Resources: k8sv1.ResourceRequirements{
@@ -165,7 +290,7 @@ func NewDataVolumeWithHTTPImportToBlockPV(dataVolumeName string, size string, ht
 			Name: dataVolumeName,
 		},
 		Spec: cdiv1.DataVolumeSpec{
-			Source: cdiv1.DataVolumeSource{
+			Source: &cdiv1.DataVolumeSource{
 				HTTP: &cdiv1.DataVolumeSourceHTTP{
 					URL: httpURL,
 				},
@@ -193,7 +318,7 @@ func NewDataVolumeCloneToBlockPV(dataVolumeName string, size string, srcNamespac
 			Name: dataVolumeName,
 		},
 		Spec: cdiv1.DataVolumeSpec{
-			Source: cdiv1.DataVolumeSource{
+			Source: &cdiv1.DataVolumeSource{
 				PVC: &cdiv1.DataVolumeSourcePVC{
 					Name:      srcName,
 					Namespace: srcNamespace,
@@ -222,7 +347,7 @@ func NewDataVolumeForUpload(dataVolumeName string, size string) *cdiv1.DataVolum
 			Annotations: map[string]string{},
 		},
 		Spec: cdiv1.DataVolumeSpec{
-			Source: cdiv1.DataVolumeSource{
+			Source: &cdiv1.DataVolumeSource{
 				Upload: &cdiv1.DataVolumeSourceUpload{},
 			},
 			PVC: &k8sv1.PersistentVolumeClaimSpec{
@@ -245,7 +370,7 @@ func NewDataVolumeForBlankRawImage(dataVolumeName, size string) *cdiv1.DataVolum
 			Annotations: map[string]string{},
 		},
 		Spec: cdiv1.DataVolumeSpec{
-			Source: cdiv1.DataVolumeSource{
+			Source: &cdiv1.DataVolumeSource{
 				Blank: &cdiv1.DataVolumeBlankImage{},
 			},
 			PVC: &k8sv1.PersistentVolumeClaimSpec{
@@ -268,7 +393,7 @@ func NewDataVolumeForBlankRawImageBlock(dataVolumeName, size string, storageClas
 			Name: dataVolumeName,
 		},
 		Spec: cdiv1.DataVolumeSpec{
-			Source: cdiv1.DataVolumeSource{
+			Source: &cdiv1.DataVolumeSource{
 				Blank: &cdiv1.DataVolumeBlankImage{},
 			},
 			PVC: &k8sv1.PersistentVolumeClaimSpec{
@@ -294,7 +419,7 @@ func NewDataVolumeForImageCloning(dataVolumeName, size, namespace, pvcName strin
 			Annotations: map[string]string{},
 		},
 		Spec: cdiv1.DataVolumeSpec{
-			Source: cdiv1.DataVolumeSource{
+			Source: &cdiv1.DataVolumeSource{
 				PVC: &cdiv1.DataVolumeSourcePVC{
 					Namespace: namespace,
 					Name:      pvcName,
@@ -319,16 +444,52 @@ func NewDataVolumeForImageCloning(dataVolumeName, size, namespace, pvcName strin
 	return dv
 }
 
+// NewDataVolumeForImageCloningAndStorageSpec initializes a DataVolume struct with spec.storage for cloning disk image
+func NewDataVolumeForImageCloningAndStorageSpec(dataVolumeName, size, namespace, pvcName string, storageClassName *string, volumeMode *k8sv1.PersistentVolumeMode) *cdiv1.DataVolume {
+	storageSpec := &cdiv1.StorageSpec{
+		AccessModes: []k8sv1.PersistentVolumeAccessMode{k8sv1.ReadWriteOnce},
+		Resources: k8sv1.ResourceRequirements{
+			Requests: k8sv1.ResourceList{
+				k8sv1.ResourceStorage: resource.MustParse(size),
+			},
+		},
+	}
+
+	dv := &cdiv1.DataVolume{
+		ObjectMeta: metav1.ObjectMeta{
+			Name:        dataVolumeName,
+			Annotations: map[string]string{},
+		},
+		Spec: cdiv1.DataVolumeSpec{
+			Source: &cdiv1.DataVolumeSource{
+				PVC: &cdiv1.DataVolumeSourcePVC{
+					Namespace: namespace,
+					Name:      pvcName,
+				},
+			},
+			Storage: storageSpec,
+		},
+	}
+	if volumeMode != nil {
+		dv.Spec.Storage.VolumeMode = volumeMode
+	}
+	if storageClassName != nil {
+		dv.Spec.Storage.StorageClassName = storageClassName
+	}
+	return dv
+}
+
 // NewDataVolumeWithRegistryImport initializes a DataVolume struct with registry annotations
 func NewDataVolumeWithRegistryImport(dataVolumeName string, size string, registryURL string) *cdiv1.DataVolume {
 	return &cdiv1.DataVolume{
 		ObjectMeta: metav1.ObjectMeta{
-			Name: dataVolumeName,
+			Name:        dataVolumeName,
+			Annotations: map[string]string{},
 		},
 		Spec: cdiv1.DataVolumeSpec{
-			Source: cdiv1.DataVolumeSource{
+			Source: &cdiv1.DataVolumeSource{
 				Registry: &cdiv1.DataVolumeSourceRegistry{
-					URL: registryURL,
+					URL: &registryURL,
 				},
 			},
 			PVC: &k8sv1.PersistentVolumeClaimSpec{
@@ -343,8 +504,8 @@ func NewDataVolumeWithRegistryImport(dataVolumeName string, size string, registr
 	}
 }
 
-// ModifyDataVolumeWithVDDKImportToBlockPV modifies a DataVolume struct (created by NewDataVolumeWithVddkImport) for importing disks from vCenter/ESX to a block PV
-func ModifyDataVolumeWithVDDKImportToBlockPV(dataVolume *cdiv1.DataVolume, storageClassName string) *cdiv1.DataVolume {
+// ModifyDataVolumeWithImportToBlockPV modifies a DataVolume struct for import to a block PV
+func ModifyDataVolumeWithImportToBlockPV(dataVolume *cdiv1.DataVolume, storageClassName string) *cdiv1.DataVolume {
 	volumeMode := corev1.PersistentVolumeMode(corev1.PersistentVolumeBlock)
 	dataVolume.Spec.PVC.VolumeMode = &volumeMode
 	dataVolume.Spec.PVC.StorageClassName = &storageClassName
@@ -358,7 +519,7 @@ func NewDataVolumeWithVddkImport(dataVolumeName string, size string, backingFile
 			Name: dataVolumeName,
 		},
 		Spec: cdiv1.DataVolumeSpec{
-			Source: cdiv1.DataVolumeSource{
+			Source: &cdiv1.DataVolumeSource{
 				VDDK: &cdiv1.DataVolumeSourceVDDK{
 					BackingFile: backingFile,
 					SecretRef:   secretRef,
@@ -386,7 +547,7 @@ func NewDataVolumeWithVddkWarmImport(dataVolumeName string, size string, backing
 			Name: dataVolumeName,
 		},
 		Spec: cdiv1.DataVolumeSpec{
-			Source: cdiv1.DataVolumeSource{
+			Source: &cdiv1.DataVolumeSource{
 				VDDK: &cdiv1.DataVolumeSourceVDDK{
 					BackingFile: backingFile,
 					SecretRef:   secretRef,
@@ -419,15 +580,18 @@ func WaitForDataVolumePhase(clientSet *cdiclientset.Clientset, namespace string,
 
 // WaitForDataVolumePhaseWithTimeout waits for DV's phase to be in a particular phase (Pending, Bound, or Lost) with a specified timeout
 func WaitForDataVolumePhaseWithTimeout(clientSet *cdiclientset.Clientset, namespace string, phase cdiv1.DataVolumePhase, dataVolumeName string, timeout time.Duration) error {
+	var actualPhase cdiv1.DataVolumePhase
+
 	err := wait.PollImmediate(dataVolumePollInterval, timeout, func() (bool, error) {
 		dataVolume, err := clientSet.CdiV1beta1().DataVolumes(namespace).Get(context.TODO(), dataVolumeName, metav1.GetOptions{})
 		if err != nil || dataVolume.Status.Phase != phase {
+			actualPhase = dataVolume.Status.Phase
 			return false, err
 		}
 		return true, nil
 	})
 	if err != nil {
-		return fmt.Errorf("DataVolume %s not in phase %s within %v", dataVolumeName, phase, timeout)
+		return fmt.Errorf("DataVolume %s not in phase %s within %v, actual phase=%s", dataVolumeName, phase, timeout, actualPhase)
 	}
 	return nil
 }
@@ -439,7 +603,7 @@ func NewDataVolumeWithArchiveContent(dataVolumeName string, size string, httpURL
 			Name: dataVolumeName,
 		},
 		Spec: cdiv1.DataVolumeSpec{
-			Source: cdiv1.DataVolumeSource{
+			Source: &cdiv1.DataVolumeSource{
 				HTTP: &cdiv1.DataVolumeSourceHTTP{
 					URL: httpURL,
 				},
@@ -474,3 +638,18 @@ func PersistentVolumeClaimFromDataVolume(datavolume *cdiv1.DataVolume) *corev1.P
 		Spec: *datavolume.Spec.PVC,
 	}
 }
+
+// GetCloneType returnc the CDI clone type
+func GetCloneType(clientSet *cdiclientset.Clientset, dataVolume *cdiv1.DataVolume) string {
+	var cloneType string
+	gomega.Eventually(func() bool {
+		dv, err := clientSet.CdiV1beta1().DataVolumes(dataVolume.Namespace).Get(context.TODO(), dataVolume.Name, metav1.GetOptions{})
+		gomega.Expect(err).ToNot(gomega.HaveOccurred())
+		val, ok := dv.Annotations["cdi.kubevirt.io/cloneType"]
+		if ok {
+			cloneType = val
+		}
+		return ok
+	}, 90*time.Second, 2*time.Second).Should(gomega.BeTrue())
+	return cloneType
+}
